/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wF0El.h"
using namespace tbox;
using namespace flow;

/**
 * @brief Evaluate the nonlinear density (constant over element)
 * @authors Adrien Crovato
 */
double F0ElRho::eval(Element &e, std::vector<double> const &u, size_t k) const
{
    double gradU2 = e.computeGrad2(u, k); // velocity squared
    double gradU = sqrt(gradU2);          // norm of velocity

    double gradUC = sqrt(mC2 * (1 / (mInf * mInf) + (gamma - 1) / 2) / (1 + (gamma - 1) / 2 * mC2)); // critical velocity
    double alpha = 1 + (gamma - 1) / 2 * mInf * mInf * (1 - gradUC * gradUC);
    double beta = mInf * mInf * gradUC / alpha;

    if (gradU < gradUC) // true density
    {
        // particularized: pow(1 + 0.5 * (gamma - 1) * mInf * mInf * (1 - gradU2), 1 / (gamma - 1));
        double a = 1 + 0.5 * (gamma - 1) * mInf * mInf * (1 - gradU2);
        return sqrt(a * a * a * a * a);
    }
    else // limited density
        return pow(alpha, 1 / (gamma - 1)) / (1 + beta * (gradU / gradUC - 1));
}
/**
 * @brief Evaluate the nonlinear density derivative (constant over element)
 * @authors Adrien Crovato
 */
double F0ElRho::evalD(Element &e, std::vector<double> const &u, size_t k) const
{
    double rho = eval(e, u, k);
    return -mInf * mInf * pow(rho, 2 - gamma); // has to be multiplied by grad u to recover true derivative
}
void F0ElRho::write(std::ostream &out) const
{
    out << "F0ElRho";
}

/**
 * @brief Evaluate the linear density (constant over element)
 * @authors Adrien Crovato
 */
double F0ElRhoL::eval(Element &e, std::vector<double> const &u, size_t k) const
{
    return 1.;
}
/**
 * @brief Evaluate the linear density derivative (constant over element)
 * @authors Adrien Crovato
 */
double F0ElRhoL::evalD(Element &e, std::vector<double> const &u, size_t k) const
{
    return 0.;
}
void F0ElRhoL::write(std::ostream &out) const
{
    out << "F0ElRhoL";
}

/**
 * @brief Evaluate the nonlinear Mach number (constant over element)
 * @authors Adrien Crovato
 */
double F0ElMach::eval(Element &e, std::vector<double> const &u, size_t k) const
{
    double gradU2 = e.computeGrad2(u, k);                             // velocity squared
    double a2 = 1 / (mInf * mInf) + 0.5 * (gamma - 1) * (1 - gradU2); // speed of sound squared

    return sqrt(gradU2) / sqrt(a2);
}
/**
 * @brief Evaluate the nonlinear Mach number derivative (constant over element)
 * @authors Adrien Crovato
 */
double F0ElMach::evalD(Element &e, std::vector<double> const &u, size_t k) const
{
    double gradU2 = e.computeGrad2(u, k);
    double a2 = 1 / (mInf * mInf) + 0.5 * (gamma - 1) * (1 - gradU2);

    return 1 / sqrt(gradU2 * a2) + 0.5 * (gamma - 1) * sqrt(gradU2) / sqrt(a2 * a2 * a2); // has to be multiplied by grad u to recover true derivative
}
void F0ElMach::write(std::ostream &out) const
{
    out << "F0ElMach";
}

/**
 * @brief Evaluate the linear Mach number (constant over element)
 * @authors Adrien Crovato
 */
double F0ElMachL::eval(Element &e, std::vector<double> const &u, size_t k) const
{
    return 0.;
}
/**
 * @brief Evaluate the linear Mach number derivative (constant over element)
 * @authors Adrien Crovato
 */
double F0ElMachL::evalD(Element &e, std::vector<double> const &u, size_t k) const
{
    return 0.;
}
void F0ElMachL::write(std::ostream &out) const
{
    out << "F0ElMachL";
}

/**
 * @brief Evaluate the nonlinear pressure coefficient (constant over element)
 * @authors Adrien Crovato
 */
double F0ElCp::eval(Element &e, std::vector<double> const &u, size_t k) const
{
    double gradU2 = e.computeGrad2(u, k); // velocity squared
    //particularized: 2 / (gamma * mInf * mInf) * (pow(1 + 0.5 * (gamma - 1) * mInf * mInf * (1 - gradU2), gamma / (gamma - 1)) - 1);
    double a = 1 + 0.5 * (gamma - 1) * mInf * mInf * (1 - gradU2);

    return 2 / (gamma * mInf * mInf) * (sqrt(a * a * a * a * a * a * a) - 1);
}
/**
 * @brief Evaluate the nonlinear pressure coefficient derivative (constant over element)
 * @authors Adrien Crovato
 */
double F0ElCp::evalD(Element &e, std::vector<double> const &u, size_t k) const
{
    double gradU2 = e.computeGrad2(u, k); // velocity squared
    double a = 1 + 0.5 * (gamma - 1) * mInf * mInf * (1 - gradU2);

    return -2 * sqrt(a * a * a * a * a); // has to be multiplied by grad u to recover true derivative
}
void F0ElCp::write(std::ostream &out) const
{
    out << "F0ElCp";
}

/**
 * @brief Evaluate the linear pressure coefficient (constant over element)
 * @authors Adrien Crovato
 */
double F0ElCpL::eval(Element &e, std::vector<double> const &u, size_t k) const
{
    return 1 - e.computeGrad2(u, k);
}
double F0ElCpL::evalD(Element &e, std::vector<double> const &u, size_t k) const
{
    return -2.; // has to be multiplied by grad u to recover true derivative
}
void F0ElCpL::write(std::ostream &out) const
{
    out << "F0ElCpL";
}