/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wMedium.h"
#include "wTag.h"
#include "wElement.h"
#include "wNode.h"
using namespace tbox;
using namespace flow;

Medium::Medium(std::shared_ptr<MshData> _msh, int no, Fct0 &_rho, Fct0 &_mach,
               Fct0 &_cP, F0PsPhiInf &_phiInf) : Group(_msh, no), rho(_rho),
                                                    mach(_mach), cP(_cP), phiInf(_phiInf)
{
    createMap();
    std::cout << "Fluid is " << *tag << "with " << nCnt << " nodes."
              << std::endl
              << std::endl;
}
Medium::Medium(std::shared_ptr<MshData> _msh, std::string const &name,
               Fct0 &_rho, Fct0 &_mach, Fct0 &_cP, F0PsPhiInf &_phiInf) : Group(_msh, name), rho(_rho), mach(_mach), cP(_cP), phiInf(_phiInf)
{
    createMap();
    std::cout << "Fluid is " << *tag << "with " << nCnt << " nodes."
              << std::endl
              << std::endl;
}

/**
 * @brief Create data structure
 * @authors Adrien Crovato
 */
void Medium::createMap()
{
    // Create map of elements connected to node
    for (auto e : tag->elems)
        for (auto n : e->nodes)
            neMap[n].push_back(e);
    nCnt = neMap.size();
    // Collect connected elements
    adjMap.resize(tag->elems.size());
    int i = 0;
    for (auto e : tag->elems)
    {
        // Get elements sharing at least one node
        std::vector<Element *> vAdjE;
        for (auto n : e->nodes)
            vAdjE.insert(vAdjE.end(), neMap.at(n).begin(), neMap.at(n).end());
        // Sort and remove duplicates
        std::sort(vAdjE.begin(), vAdjE.end());
        auto iterator = std::unique(vAdjE.begin(), vAdjE.end());
        vAdjE.resize(std::distance(vAdjE.begin(), iterator));
        // Remove self
        for (size_t j = 0; j < vAdjE.size(); ++j)
        {
            if (vAdjE[j] == e)
                vAdjE.erase(vAdjE.begin() + j);
        }
        // Pair an element and its adjacent elements
        adjMap[i] = std::pair<Element *, std::vector<Element *>>(e, vAdjE);
        i++;
    }
}

void Medium::write(std::ostream &out) const
{
    out << "Fluid on " << *tag << " with:\n\t"
        << "rho = " << this->rho << ", mach = " << this->mach << ", cP = " << this->cP << "\n";
    out << "Adjencency map:\n";
    for (auto p : adjMap)
    {
        out << "\t" << p.first->no << ':';
        for (size_t j = 0; j < p.second.size(); j++)
            out << p.second[j]->no << ',';
        out << ";\n";
    }
}
