#! /usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# @brief Compute flow around the Agard445 wing at 1 degrees AOA and Mach 0.80
# @authors Adrien Crovato

from __future__ import print_function
import numpy as np
import flow as flo
import flow.utils as floU
import flow.default as floD
import tbox
import tbox.utils as tbxU
import tbox.gmsh as gmsh
import fwk
from fwk.testing import *
from fwk.coloring import ccolors

try:   
   import tboxVtk
   BestWriter = tboxVtk.VtkExport
   canPost = True
except:
   BestWriter = tbox.GmshExport
   canPost = False

def main():
	# timer
    tms = fwk.Timers()
    tms['total'].start()

    # define flow variables
    alpha = 1*np.pi/180
    U_inf = [np.cos(alpha), 0., np.sin(alpha)] # norm should be 1
    M_inf = 0.8
    M_crit = 5 # Squared critical Mach number (above which density is modified)
    dim = len(U_inf)

    # define dimension and mesh size
    lgt = 7. # box length
    hgt = 6. # box height
    wdt = 3. # box width
    xO = -3. # box x-origin
    zO = -3. # box z-origin
    S_ref = .35 # reference area
    c_ref = .47 # reference chord (MAC)
    rlems = 0.0028 # root leading edge mesh size
    rtems = 0.0056 # root trailing edge mesh size
    tlems = 0.0018 # tip leading mesh size
    ttems = 0.0045 # tip trailing mesh size
    fms = 1.0 # farfield mesh size

    # mesh the wing
    print(ccolors.ANSI_BLUE + 'PyMeshing...' + ccolors.ANSI_RESET)
    tms['msh'].start()
    pars = {'xL' : lgt, 'yL' : wdt, 'zL' : hgt, 'xO' : xO, 'zO' : zO, 'msLeRt' : rlems, 'msTeRt' : rtems, 'msLeTp' : tlems, 'msTeTp' : ttems, 'msF' : fms}
    msh = gmsh.MeshLoader("../models/agard445.geo",__file__).execute(**pars)
    # crack the mesh
    mshCrck = tbox.MshCrack(msh, dim)
    mshCrck.setCrack('wake')
    mshCrck.addBoundaries(['field', 'wing', 'symmetry', 'downstream'])
    mshCrck.setExcluded('wakeTip')
    mshCrck.run()
    del mshCrck
    tbox.GmshExport(msh).save(msh.name)
    writer = BestWriter(msh)
    tms['msh'].stop()

    # set the problem and add medium, initial/boundary conditions
    tms['pre'].start()
    pbl = flo.Problem(msh, dim, alpha, M_inf, S_ref, c_ref, 0., 0., 0.)
    pbl.set(flo.Medium(msh, 'field', flo.F0ElRho(M_inf, M_crit), flo.F0ElMach(M_inf), flo.F0ElCp(M_inf), flo.F0PsPhiInf(dim, alpha)))
    pbl.add(flo.Initial(msh, 'field', flo.F0PsPhiInf(dim, alpha)))
    pbl.add(flo.Dirichlet(msh, 'upstream', flo.F0PsPhiInf(dim, alpha)))
    pbl.add(flo.Freestream(msh, 'farfield', tbox.Fct1C(-U_inf[0], -U_inf[1], -U_inf[2])))
    pbl.add(flo.Freestream(msh, 'downstream', tbox.Fct1C(-U_inf[0], -U_inf[1], -U_inf[2])))
    pbl.add(flo.Wake(msh, ['wake', 'wake_', 'field', 'teTip'])) 
    pbl.add(flo.Boundary(msh, ['wing', 'field']))
    tms['pre'].stop()

    # solve problem
    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver'].start()
    solver = floD.newton(pbl)
    solver.run()
    solver.save(0, writer)
    tms['solver'].stop()

    # post process
    tms['post'].start()
    if canPost:
        floU.writeSlices(msh.name,[0.01, 0.37, 0.75],5)
        data = tbxU.read('slice_1.dat')
        tbxU.plot(data[:,3], data[:,4], 'x', 'Cp', 'Pressure distribution at MAC', True)
    tms['post'].stop()

    # display timers
    tms['total'].stop()
    print(ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET)
    print('CPU statistics')
    print(tms)

    # check results
    print(ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET)
    tests = CTests()
    tests.add(CTest('CL', solver.Cl, 0.062, 1e-1))
    tests.add(CTest('CD', solver.Cd, 0.0006, 1e-1))
    tests.run()

    # eof
    print('')

if __name__ == "__main__":
    main()