/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WTIME_H
#define WTIME_H

#include "fwk.h"
#include <time.h>
#include <string>
#include <iostream>

namespace fwk
{

/**
 * @brief Portable base class handling a time
 */
class FWK_API Time
{
public:
    virtual double sec() const = 0;
    std::string hms() const;
#ifndef SWIG
    friend FWK_API std::ostream &operator<<(std::ostream &out, Time const &obj);
#endif
};

//-----------------------------------------------------------------------------------------

#ifndef SWIG

#ifdef WIN32

/**
 * @brief Time as a __int64 (MS Windows only)
 */
class FWK_API Int64Time : public Time
{
    __int64 time;

public:
    explicit Int64Time(__int64 t = 0) : time(t) {}
    virtual double sec() const override;
    operator bool() const { return time != 0; }
    Int64Time operator+(Int64Time const &obj) const { return Int64Time(time + obj.time); }
    Int64Time operator-(Int64Time const &obj) const { return Int64Time(time - obj.time); }
};

#else //WIN32

class FWK_API UnixClockTime : public Time
{
    clock_t time;

public:
    explicit UnixClockTime(clock_t t = 0) : time(t) {}
    virtual double sec() const override;
    operator bool() const { return time != 0; }
    UnixClockTime operator+(UnixClockTime const &obj) const { return UnixClockTime(time + obj.time); }
    UnixClockTime operator-(UnixClockTime const &obj) const { return UnixClockTime(time - obj.time); }
};

#endif //WIN32

//-----------------------------------------------------------------------------------------

/**
 * @brief Time as a clock_t (freq = CLOCKS_PER_SEC)
 */
class FWK_API ClockTime : public Time
{
    clock_t time;

public:
    explicit ClockTime(clock_t t = 0) : time(t) {}
    virtual double sec() const override;
    operator bool() const { return time != 0; }
    ClockTime operator+(ClockTime const &obj) const { return ClockTime(time + obj.time); }
    ClockTime operator-(ClockTime const &obj) const { return ClockTime(time - obj.time); }
};

//-----------------------------------------------------------------------------------------

/**
 * @brief Time as a double
 */
class FWK_API DblTime : public Time
{
    double time;

public:
    explicit DblTime(double t = 0.0) : time(t) {}
    virtual double sec() const override;
    operator bool() const { return time != 0.0; }
    DblTime operator+(DblTime const &obj) const { return DblTime(time + obj.time); }
    DblTime operator-(DblTime const &obj) const { return DblTime(time - obj.time); }
};

#endif //SWIG

};     // namespace fwk
#endif //WTIME_H
