#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# simple test on a square using 'heat'

import heat as h
import heat.utils as u
import tbox
import tbox.gmsh as gmsh
from fwk.testing import *
import math

class MySource(tbox.Fct0XYZ):
    """
    Custom heat source
    """
    def __init__(self, cx, cy, freq):
        tbox.Fct0XYZ.__init__(self)
        self.freq = freq
        self.cx   = cx
        self.cy   = cy
    def eval(self, pt):
        x = pt[0]-self.cx
        y = pt[1]-self.cy
        return math.sin(2*math.pi*self.freq*math.sqrt(x*x+y*y)/1.)

def main():

    # mesh a unit square
    pars={ 'Lx' : 1 , 'Ly' : 1, 'Nx' : 50, 'Ny' : 50}
    msh = gmsh.MeshLoader("../../models/rect_stru.geo",__file__).execute(**pars)
    gmshWriter = tbox.GmshExport(msh)
    pbl = h.Problem(msh)

    # source calculated by python using swig directors
    pbl.add(h.Source(msh, "S1", MySource(0.2,0.3, 5.0)))

    # medium/boundary conditions
    pbl.add(h.Medium(msh, "S1", tbox.Fct2C(0.1, 0.1, 0.0)))
    pbl.add(h.Dirichlet(msh,"Left", 0.))
    pbl.add(h.Dirichlet(msh,"Right", 0.))
    pbl.add(h.Dirichlet(msh,"Bottom", 0.))
    pbl.add(h.Dirichlet(msh,"Top", 0.))

    # solve problem
    solver = u.solve(pbl, gmshWriter)

    # extract some results
    x1,y1,T1 = u.getTemp(solver,"S1", sort='y')
    
    u.simpleplot(y1, T1, xlabel='y [mm]', title='Temperature along y')
       
    tests = CTests()
    tests.add(CTest('max(T)', max(T1), 0.014730, 1e-3))
    tests.run()


if __name__ == "__main__":
    main()
