#!/usr/bin/env python
# -*- coding: latin-1; -*-

from __future__ import print_function
from __future__ import division
from builtins import str
from builtins import range
from past.utils import old_div
import katropton as m
import tbox
import tbox.gmsh as gmsh
import katropton.utilities as u
import numpy as np
import shlex
import subprocess
import os
import tarfile
import glob

from katropton.Juelich.UQ.halton import *
from katropton.Juelich.UQ.copulas import *


def irradiation(irradotion_index, p1=1., p3=1.95):
    sigma_1_min = 0.1
    sigma_1_max = 0.1
    sigma_3_min = 0.05
    sigma_3_max = 0.05

    sigma_1 = sigma_1_min + (sigma_1_max - sigma_1_min) * irradotion_index
    sigma_3 = sigma_3_min + (sigma_3_max - sigma_3_min) * irradotion_index

    mode_min = 70
    mode_max = 180

    mode_1 = mode_min + (mode_max - mode_min) * irradotion_index**p1
    mode_3 = mode_min + (mode_max - mode_min) * irradotion_index**p3

    mu_1 = np.log(mode_1)  # + sigma_1**2
    mu_3 = np.log(mode_3)  # + sigma_3**2

    return sigma_1, sigma_3, mode_1, mode_3, mu_1, mu_3


def random_case(N, id_case):
    comm, rank, siz = m.utilities.import_MPI()
    np.random.seed(42)

    sigma = 0.15
    mode = 180
    mu = np.log(mode) + sigma**2

    qMC = True

    if id_case == 'qMC 1':
        n_rand = 2
        halton_points = np.zeros((n_rand, N))
        for i in range(0, N):
            halton_points[:, i] = halton(i + 1, n_rand)

        randomVariable_halton = lognormal(halton_points, [mu, mu],
                                          [sigma, sigma])
        randomVariable_1 = randomVariable_halton[0, :]
        randomVariable_2 = randomVariable_halton[1, :]
        randomVariable_3 = mode * np.ones((N, ))
    elif id_case == 'MC 1':
        randomVariable_1 = np.random.lognormal(mu, sigma, N)
        randomVariable_2 = np.random.lognormal(mu, sigma, N)
        randomVariable_3 = mode * np.ones((N, ))
    elif id_case == 'qMC 2':
        mode_min = 70
        mode_max = 180
        n_rand = 2
        halton_points = np.zeros((n_rand, N))
        for i in range(0, N):
            halton_points[:, i] = halton(i, n_rand)

        randomVariable_halton = lognormal(halton_points, [mu, mu],
                                          [sigma, sigma])
        randomVariable_1 = mode_min + (mode_max -
                                       mode_min) * halton_points[0, :]
        randomVariable_2 = mode_min + (mode_max -
                                       mode_min) * halton_points[0, :]
        randomVariable_3 = mode_min + (mode_max -
                                       mode_min) * halton_points[1, :]
    elif id_case == 'qMC 3':
        from scipy.stats import lognorm
        mode_min = 70
        mode_max = 180
        n_rand = 3
        halton_points = np.zeros((n_rand, N))

        randomVariable_1 = np.zeros((N, ))
        randomVariable_3 = np.zeros((N, ))

        for i in range(0, N):
            halton_points[:, i] = halton(i + 1, n_rand)
            sigma_1, sigma_3, mode_1, mode_3, mu_1, mu_3 = irradiation(
                halton_points[0, i])
            randomVariable_1[i] = lognorm.ppf(halton_points[1, i],
                                              sigma_1,
                                              scale=np.exp(mu_1))
            randomVariable_3[i] = lognorm.ppf(halton_points[2, i],
                                              sigma_3,
                                              scale=np.exp(mu_3))

        randomVariable_2 = randomVariable_1
    elif id_case == 'Xiao':
        file_dir = os.path.dirname(__file__)
        loaded = np.load(file_dir + '/../samples_files/Xiao_' + str(N) +
                         '.npz')
        samples = loaded['samples']
        N = samples.shape[0]
        weights = loaded['weights']
        randomVariable_1 = samples[0:N, 0]
        randomVariable_3 = samples[0:N, 1]
        randomVariable_2 = randomVariable_1

    if rank == 0:
        np.savetxt('randomVariable_1.txt', randomVariable_1)
        np.savetxt('randomVariable_2.txt', randomVariable_2)
        np.savetxt('randomVariable_3.txt', randomVariable_3)

    return randomVariable_1, randomVariable_2, randomVariable_3


def material_option(option):
    return int(option[-1])


def clamping_option(option):
    return str(option[0])


def scale_msh(msh, scale):
    for n in msh.nodes:
        for i in range(0, 3):
            n.pos[i] = n.pos[i] * scale


def tar_results(directory, filename, rank, all=False):
    os.chdir(directory)
    if rank == 0:
        tf = tarfile.open(filename + ".tar.gz", mode="w:gz")
        if all:
            for file in glob.glob('*'):
                if file.endswith(".tar.gz"):
                    print(file + ' is not included')
                else:
                    tf.add(file)
        else:
            for file in glob.glob('r*.txt'):
                tf.add(file)
            for file in glob.glob('*/*/t*.txt'):
                tf.add(file)
            for file in glob.glob('*/*/b*.txt'):
                tf.add(file)
            for file in glob.glob('*/*/*.npz'):
                tf.add(file)
            for file in glob.glob('*/*/*.out'):
                tf.add(file)
        tf.close()


def evaluate_one_ensemble(ensemble_size, msh, comm, file_dir, randomVariable_1,
                          randomVariable_2, randomVariable_3, scale_u,
                          scale_T, option, baking=False):

    from fwk.wutils import parseargs
    args = parseargs()

    from PyTrilinos import Teuchos
    import numpy as np

    small_disk_footprint = False
    use_block = True

    comm, rank, siz = m.utilities.import_MPI()

    solverList = Teuchos.ParameterList()
    solverList['type'] = "BlockGmres"

    solverList['Maximum Iterations'] = 500
    solverList['Verbosity'] = 33
    solverList['Flexible Gmres'] = False
    solverList['Output Style'] = 1
    solverList['Output Frequency'] = 1

    solverList['Use preconditioner'] = True

    solverList['Print Teuchos timers'] = True
    solverList['Write txt files'] = True
    if small_disk_footprint:
        solverList['Write vtk files'] = False

    solverList['convert MueLu xml file'] = True
    solverList["use xml file"] = True

    if use_block:
        if ensemble_size >= 16 and siz == 2 and clamping_option(option) == "A":
            solverList["Num Blocks"] = 150
        else:
            solverList["Num Blocks"] = 200

        solverList['Use blocked matrix'] = True

        solverList['Use blocked status test'] = True
        solverList['Use blocked status test: x'] = True
        solverList['Use blocked status test: T'] = True
        solverList['Use weighted status test'] = True

        solverList['Ensemble Convergence Tolerance: T - relative'] = True
        solverList['Ensemble Convergence Tolerance: x - relative'] = True
        solverList[
            'Ensemble Convergence Tolerance: T - relative full rhs'] = False
        solverList[
            'Ensemble Convergence Tolerance: x - relative full rhs'] = False
        solverList[
            'Ensemble Convergence Tolerance: weights - relative'] = False

        solverList['Ensemble Convergence Tolerance'] = 10**(-5)
        solverList['Ensemble Convergence Tolerance: x'] = 10**(-6)
        solverList['Ensemble Convergence Tolerance: T'] = 10**(-7)
        solverList['Ensemble Convergence Tolerance: weights'] = 10**(-3)

        solverList[
            "MueLu xml file name"] = file_dir + '/../smoothers/AMG_BGS_gs_gs.xml'
    else:
        solverList['Ensemble Convergence Tolerance'] = 10**(-8)
        solverList[
            "MueLu xml file name"] = file_dir + '/../smoothers/heat_no_contact_mtgs.xml'

        solverList["Scaled Block 0"] = True
    pbl = m.Problem(msh, comm)

    # Distance in the mesh are represented in meters

    scale_E = 1. / scale_u**2
    scale_k = old_div(scale_T, scale_u)
    scale_hsource = 1. / scale_u**3
    scale_hflux = 1. / scale_u**2
    scale_d = scale_T

    # 379 000 MPa = 379 GPa = 379E9 Pa
    E_rhodium = 379E9 * np.ones(ensemble_size) * scale_E
    nu_rhodium = 0.26 * np.ones(ensemble_size)
    # 0.150 kW/(m K) = 150 W/(m K)
    k_rhodium = 150 * scale_k
    d_rhodium = 0.781E-05 * scale_d

    E_WCu = 280E9 * np.ones(ensemble_size) * scale_E
    nu_WCu = 0.298 * np.ones(ensemble_size)
    k_WCu = 180 * scale_k
    d_WCu = 0.88E-05 * scale_d

    E_SS = 200E9 * np.ones(ensemble_size) * scale_E
    nu_SS = 0.3 * np.ones(ensemble_size)
    k_SS = 15.3 * scale_k
    d_SS = 1.57E-05 * scale_d

    E_CuCrZr = 118E9 * np.ones(ensemble_size) * scale_E
    nu_CuCrZr = 0.33 * np.ones(ensemble_size)
    k_CuCrZr = 345 * np.ones(ensemble_size) * scale_k
    d_CuCrZr = 1.8E-05 * np.ones(ensemble_size) * scale_d

    E_AlN = 320E9 * np.ones(ensemble_size) * scale_E
    nu_AlN = 0.24 * np.ones(ensemble_size)
    k_AlN = 180 * np.ones(ensemble_size) * scale_k
    d_AlN = 0.48E-05 * np.ones(ensemble_size) * scale_d

    E_I718 = 183E9 * np.ones(ensemble_size) * scale_E
    nu_I718 = 0.31 * np.ones(ensemble_size)
    k_I718 = 15.8 * scale_k
    d_I718 = 1.38E-05 * scale_d

    T_ref = 22.

    m.Medium(pbl, "Mirror", "rhodium", E_rhodium, nu_rhodium, k_rhodium,
             d_rhodium)
    m.Medium(pbl, "Substrate", "WCu", E_WCu, nu_WCu, k_WCu, d_WCu)

    if material_option(option) == 1:
        m.Medium(pbl, "Holder", "SS", E_SS, nu_SS, k_SS, d_SS)
    elif material_option(option) == 2:
        m.Medium(pbl, "Holder", "CuCrZr", E_CuCrZr, nu_CuCrZr, k_CuCrZr,
                 d_CuCrZr)

    m.Medium(pbl, "Washer 1", "I718", E_I718, nu_I718, k_I718, d_I718)
    m.Medium(pbl, "Washer 2", "I718", E_I718, nu_I718, k_I718, d_I718)
    m.Medium(pbl, "Washer 3", "I718", E_I718, nu_I718, k_I718, d_I718)

    m.Medium(pbl, "Shaft 1", "I718", E_I718, nu_I718, k_I718, d_I718)
    m.Medium(pbl, "Shaft 2", "I718", E_I718, nu_I718, k_I718, d_I718)
    m.Medium(pbl, "Shaft 3", "I718", E_I718, nu_I718, k_I718, d_I718)

    m.Medium(pbl, "Nut 1", "I718", E_I718, nu_I718, k_I718, d_I718)
    m.Medium(pbl, "Nut 2", "I718", E_I718, nu_I718, k_I718, d_I718)
    m.Medium(pbl, "Nut 3", "I718", E_I718, nu_I718, k_I718, d_I718)

    m.Medium(pbl, "Spacer 1 part 1", "AlN", E_AlN, nu_AlN,
             randomVariable_1 * scale_k, d_AlN)
    m.Medium(pbl, "Spacer 2 part 1", "AlN", E_AlN, nu_AlN,
             randomVariable_2 * scale_k, d_AlN)
    m.Medium(pbl, "Spacer 3 part 1", "AlN", E_AlN, nu_AlN,
             randomVariable_3 * scale_k, d_AlN)

    m.Medium(pbl, "Spacer 1 part 2", "WCu", E_WCu, nu_WCu, k_WCu, d_WCu)
    m.Medium(pbl, "Spacer 2 part 2", "WCu", E_WCu, nu_WCu, k_WCu, d_WCu)
    m.Medium(pbl, "Spacer 3 part 2", "WCu", E_WCu, nu_WCu, k_WCu, d_WCu)

    if clamping_option(option) == "A":
        m.Dirichlet(pbl, "Holder z clamping", "Clamped", 0, 0., 0, 0., 1, 0.,
                    0, 0., ensemble_size)
        m.Dirichlet(pbl, "Holder x-y clamping", "Clamped", 1, 0., 1, 0., 0,
                    0., 0, 0., ensemble_size)

    elif clamping_option(option) == "B":
        m.Dirichlet(pbl, "Holder x-y-z clamping point", "Clamped", 1, 0., 1,
                    0., 1, 0., 0, 0., ensemble_size)
        m.Dirichlet(pbl, "Holder z clamping point", "Clamped", 0, 0., 0, 0.,
                    1, 0., 0, 0., ensemble_size)
        m.Dirichlet(pbl, "Holder z-y clamping point", "Clamped", 0, 0., 1, 0.,
                    1, 0., 0, 0., ensemble_size)

    m.Dirichlet(pbl, "Cooling channels", "Clamped", 0, 0., 0, 0., 0, 0., 1,
                old_div((70. - T_ref), scale_T), ensemble_size)

    if not baking:
        zero = np.zeros(ensemble_size)
        hf = 20. / 0.0121 * np.ones(
            ensemble_size) * scale_hflux  # 20 W / (0.0121 m^2)

        m.Neumann(pbl, "Mirror surface", "heat fluxes", 0, zero, 0, zero, 0, zero,
                  1, hf, ensemble_size)

        m.Source(pbl, "Mirror", "ihg", 20E6 * scale_hsource)
        m.Source(pbl, "Substrate", "ihg", 2.45E6 * scale_hsource)

        if material_option(option) == 1:
            m.Source(pbl, "Holder", "ihg", 0.7E6 * scale_hsource)
        elif material_option(option) == 2:
            m.Source(pbl, "Holder", "ihg", 0.6E6 * scale_hsource)

        m.Source(pbl, "Spacer 1 part 1", "ihg", 0.3E6 * scale_hsource)
        m.Source(pbl, "Spacer 2 part 1", "ihg", 0.3E6 * scale_hsource)
        m.Source(pbl, "Spacer 3 part 1", "ihg", 0.3E6 * scale_hsource)

        m.Source(pbl, "Spacer 1 part 2", "ihg", 2.45E6 * scale_hsource)
        m.Source(pbl, "Spacer 2 part 2", "ihg", 2.45E6 * scale_hsource)
        m.Source(pbl, "Spacer 3 part 2", "ihg", 2.45E6 * scale_hsource)

        m.Source(pbl, "Shaft 1", "ihg", 0.8E6 * scale_hsource)
        m.Source(pbl, "Shaft 2", "ihg", 0.8E6 * scale_hsource)
        m.Source(pbl, "Shaft 3", "ihg", 0.8E6 * scale_hsource)

        m.Source(pbl, "Nut 1", "ihg", 0.8E6 * scale_hsource)
        m.Source(pbl, "Nut 2", "ihg", 0.8E6 * scale_hsource)
        m.Source(pbl, "Nut 3", "ihg", 0.8E6 * scale_hsource)

        m.Source(pbl, "Washer 1", "ihg", 0.8E6 * scale_hsource)
        m.Source(pbl, "Washer 2", "ihg", 0.8E6 * scale_hsource)
        m.Source(pbl, "Washer 3", "ihg", 0.8E6 * scale_hsource)

    if use_block:
        m.Weight(pbl, "Mirror", 1, 0., 1, 0., 1, 1., 1, 0., ensemble_size)

    slv = m.IterativeSolver(pbl, args.k, solverList, 4, ensemble_size)

    slv.start()

    if rank == 0:
        f = open('timers.txt', 'w')
        print(slv.getTimers(), file=f)
        f.closed

    tag_name = "Mirror surface"
    name_sol = "x_mm.txt"
    DOFperNode = 4
    which_dof = 2

    x, y, z, tri, dz = u.extract_2D(msh, tag_name, DOFperNode, which_dof,
                                    ensemble_size, name_sol)
    which_dof = 3
    x, y, z, tri, T = u.extract_2D(msh, tag_name, DOFperNode, which_dof,
                                   ensemble_size, name_sol)
    np.savez('mirror_data_' + str(rank), x=x, y=y, z=z, tri=tri, dz=dz, T=T)

    if rank == 0:
        for i in range(1, siz):
            data = comm.recv(source=i, tag=11)
    else:
        comm.send(1, dest=0, tag=11)

    if small_disk_footprint:
        if rank == 0:
            os.remove(name_sol)

    if rank == 0:
        #OAX = msh.ntags["Optical Axis"].elems[0].nodes[0].pos[0]
        #OAY = msh.ntags["Optical Axis"].elems[0].nodes[0].pos[1]
        OAX, OAY = 0.0595, 0.017
        OAX = old_div(OAX, scale_u)
        OAY = old_div(OAY, scale_u)
        #OAX, OAY = 0.0595, 0.017
        for i in range(0, siz):
            npzfile = np.load('mirror_data_' + str(i) + '.npz')
            x_i = npzfile['x']
            y_i = npzfile['y']
            z_i = 0. * x_i
            dz_i = npzfile['dz']
            tri_i = npzfile['tri']
            T_i = npzfile['T']
            if i == 0:
                x = x_i
                y = y_i
                z = z_i
                dz = dz_i
                tri = tri_i
                T = T_i
                nNodes = len(x)
            else:
                x = np.append(x, x_i, axis=0)
                y = np.append(y, y_i, axis=0)
                z = np.append(z, z_i, axis=0)
                dz = np.append(dz, dz_i, axis=0)
                tri = np.append(tri, tri_i + nNodes, axis=0)
                T = np.append(T, T_i, axis=0)
                nNodes = len(x)
        x = old_div(x, scale_u)
        y = old_div(y, scale_u)
        z = old_div(z, scale_u)
        dz = old_div(dz, scale_u)
        T = T * scale_T + T_ref
        np.savez('mirror_data_all', x=x, y=y, z=z, tri=tri, dz=dz, T=T)
        for i in range(0, ensemble_size):
            optical_coefficients, dz_residual = u.compute_RBM_Curvature_Irreg(
                x, y, z, tri, dz[:, i], "m", np.array([OAX, OAY]))
            np.savez('mirror_optical_results_' + str(i),
                     optical_coefficients=optical_coefficients,
                     dz_residual=dz_residual)
            np.savetxt('mirror_optical_results_' + str(i) + '.out',
                       optical_coefficients)

    if small_disk_footprint:
        if rank == 0:
            for i in range(0, siz):
                os.remove('mirror_data_' + str(i) + '.npz')


def evaluate_all_ensembles(ensemble_sizes, N, msh, comm, file_dir,
                           randomVariable1, randomVariable2, randomVariable3,
                           scale_u, scale_T, option, baking=False):
    """
    This function loop over the ensemble sizes,
    and the ensembles
    """
    for ensemble_size in ensemble_sizes:
        directory = str(ensemble_size)
        u.mkdir_MPI(directory, comm, rank, size)
        os.chdir(directory)

        N_ensemble = int(np.floor(old_div(N, ensemble_size)))

        ensemble_index_min = 0
        for ensemble_index in range(ensemble_index_min, N_ensemble):
            directory = str(ensemble_index)
            u.mkdir_MPI(directory, comm, rank, size)
            os.chdir(directory)

            ensemble_index_start = ensemble_index * ensemble_size
            ensemble_index_end = ensemble_index_start + ensemble_size

            randomVariable_cur1 = randomVariable1[ensemble_index_start:
                                                  ensemble_index_end]

            randomVariable_cur2 = randomVariable2[ensemble_index_start:
                                                  ensemble_index_end]

            randomVariable_cur3 = randomVariable3[ensemble_index_start:
                                                  ensemble_index_end]

            evaluate_one_ensemble(ensemble_size, msh, comm, file_dir,
                                  randomVariable_cur1, randomVariable_cur2,
                                  randomVariable_cur3, scale_u, scale_T, option, baking)
            os.chdir('..')
        os.chdir('..')


def main():
    comm, rank, siz = m.utilities.import_MPI()

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    directory = 'MPI_' + str(siz)
    u.mkdir_MPI(directory, comm, rank, size)
    os.chdir(directory)

    mesh_name = 'fused_mirror_assembly_mm_part_' + str(siz) + '.msh'

    msh = gmsh.MeshLoader(mesh_name, work_dir).execute(myrank=rank)

    scale_u_1 = 1000.
    scale_u_2 = 1.
    scale_u = scale_u_1 * scale_u_2
    scale_T = 1.

    scale_msh(msh, scale_u_2)

    N = 20
    id_case = 'Xiao'

    randomVariable_1, randomVariable_2, randomVariable_3 = random_case(
        N, id_case)

    ensemble_size = int(os.getenv('ensemble_size', 1))

    ensemble_sizes = [ensemble_size]

    N = len(randomVariable_1)

    u.mkdir_MPI(option, comm, rank, size)
    os.chdir(option)

    evaluate_all_ensembles(ensemble_sizes, N, msh, comm, file_dir,
                           randomVariable_1, randomVariable_2,
                           randomVariable_3, scale_u, scale_T, option, baking=False)
    os.chdir('..')

    option = 'B2'

    u.mkdir_MPI(option, comm, rank, size)
    os.chdir(option)

    evaluate_all_ensembles(ensemble_sizes, N, msh, comm, file_dir,
                           randomVariable_1, randomVariable_2,
                           randomVariable_3, scale_u, scale_T, option, baking=False)
    os.chdir('..')

    os.chdir('..')


if __name__ == "__main__":
    main()
