#ifndef KATROPTON_DISCRETEPROBLEM_H
#define KATROPTON_DISCRETEPROBLEM_H

#include "katropton.h"
#include "Map.h"
#include "Domain.h"
#include "Loads.h"
#include "Algebraic.h"
#include "ElementMatrices.h"
#include "ElementVectors.h"

#include <Xpetra_Map.hpp>
#include <Xpetra_TpetraMap.hpp>
#include <Xpetra_BlockedMap.hpp>
#include <Xpetra_BlockedCrsMatrix.hpp>
#include <Xpetra_BlockedMultiVector.hpp>
#include <Xpetra_TpetraCrsMatrix.hpp>
#include <Xpetra_TpetraMultiVector.hpp>
#include <Xpetra_CrsMatrixWrap.hpp>
//#include <Xpetra_UseShortNames.hpp>

#include "LinearSolver.h"

#include "Mortar.h"

#include <mpi.h>

namespace katropton
{

/**
 * @brief Class which is used to store all the information related to the discretized problem:
 * <ul>
 *  <li> the information related to the activity of the Lagrange multipliers,
 *  <li> the discretized domain,
 *  <li> the algebraic information.
 * </ul>
 */
template <typename scalar>
class DiscreteProblem
{
private:
    Kokkos::View<scalar *, Kokkos::LayoutRight> old_old_activity;
    Kokkos::View<scalar *, Kokkos::LayoutRight> old_activity;

public:
    //
    // Convenience typedefs
    //
    typedef Map::local_ordinal_type local_ordinal_type;
    typedef Map::global_ordinal_type global_ordinal_type;
    typedef Map::node_type node_type;
    typedef Tpetra::Export<local_ordinal_type, global_ordinal_type> export_type;

    typedef Map::tpetra_map_type tpetra_map_type;
    typedef Map::xpetra_map_type xpetra_map_type;
    typedef Map::xpetra_tmap_type xpetra_tmap_type;
    typedef Map::xpetra_bmap_type xpetra_bmap_type;

    typedef typename Vectors<scalar>::tpetra_vector_type tpetra_vector_type;
    typedef typename Vectors<scalar>::tpetra_mvector_type tpetra_mvector_type;
    typedef typename Vectors<scalar>::xpetra_bmvector_type xpetra_bmvector_type;
    typedef typename Vectors<scalar>::xpetra_mvector_type xpetra_mvector_type;
    typedef typename Vectors<scalar>::xpetra_tmvector_type xpetra_tmvector_type;

    typedef typename Matrices<scalar>::tpetra_crs_type tpetra_crs_type;
    typedef typename Matrices<scalar>::xpetra_crs_type xpetra_crs_type;
    typedef typename Matrices<scalar>::xpetra_wcrs_type xpetra_wcrs_type;
    typedef typename Matrices<scalar>::xpetra_bcrs_type xpetra_bcrs_type;
    typedef typename Matrices<scalar>::xpetra_tcrs_type xpetra_tcrs_type;
#ifndef SWIG
    Teuchos::RCP<Domain<scalar>> domain;
    Teuchos::RCP<Loads<scalar>> loads;
    Teuchos::RCP<Algebraic<scalar>> algebraic;
#endif
    size_t numPrimalDPN;

    DiscreteProblem(Problem &pbl, size_t _numPrimalDPN,
                    Teuchos::RCP<Teuchos::ParameterList> randomParams,
                    Kokkos::View<scalar *, Kokkos::LayoutLeft> m_rv_values);

    ~DiscreteProblem()
    {
        std::cout << "~DiscreteProblem()\n";
    }

#ifndef SWIG
    void createBlockMatrix(scalar gamma, bool write, bool merge);
    void createBlockMVector(bool write);
    void computeMatrices(void);
    void computeContactMatrices(MPI_Comm comm);
    void computeLoads(bool scaled = false);
    Mask<scalar> updateActivity(int active_set_iteration);

#endif
};
}; // namespace katropton

#endif //KATROPTON_DISCRETEPROBLEM_H
