#ifndef KATROPTON_SOURCESLIST_H
#define KATROPTON_SOURCESLIST_H

#include "katropton.h"

#include "Map.h"
#include "wSource.h"
#include "wTag.h"

#include <Tpetra_Map.hpp>
#include <Tpetra_Vector.hpp>
#include <Teuchos_RCP.hpp>
#include <Kokkos_ViewFactory.hpp>

namespace katropton
{

/**
 * @brief Class which is used to store the list of the volumetric heat sources.
 */
template <typename scalar>
class SourcesList
{
    typedef Map::local_ordinal_type local_ordinal_type;
    typedef Map::global_ordinal_type global_ordinal_type;

private:
    Kokkos::View<int **, Kokkos::LayoutRight> sources_elementsList;
    Kokkos::View<scalar ***, Kokkos::LayoutRight> sources_values_list;

    local_ordinal_type sources_number;
    local_ordinal_type sources_size;
    local_ordinal_type sources_elements;

    local_ordinal_type sources_time;
    local_ordinal_type sources_values;

public:
    SourcesList();
    SourcesList(Problem &pbl, Teuchos::RCP<Map> map, Teuchos::RCP<ElementsList> elementsList);

    /**
     * @brief Return the number of sources.
     */
    inline local_ordinal_type getSourceNumber() { return sources_number; }

    /**
     * @brief Return the number of elements for a
     * given source.
     * 
     * Argument:
     * <ul>
     *  <li> i: the source ID.
     * </ul>
     */
    inline local_ordinal_type getSourceSize(local_ordinal_type i)
    {
        return sources_elementsList(i, sources_size);
    }

    /**
     * @brief Return the local ID of the element e1 of 
     * a source.
     * 
     * Arguments:
     * <ul>
     *  <li> i: the source ID,
     *  <li> e1: the ID of the element on the current BC.
     * </ul>
     */
    inline local_ordinal_type getSourceElement(local_ordinal_type i, local_ordinal_type ea)
    {
        return sources_elementsList(i, sources_elements + ea);
    }

    /**
     * @brief Return the value of the source.
     * 
     * Argument:
     * <ul>
     *  <li> i: the source ID.
     * </ul> 
     */
    inline scalar getSourceValue(local_ordinal_type i)
    {
        return sources_values_list(i, 0, sources_values);
    }
};

/**
 * @brief SourcesList constructor
 * 
 * This constructor allocates 2 Kokkos::View which store all the required information related to source, including:
 * <ul>
 *  <li> all the source elements,
 *  <li> all the values of the heat source.
 * </ul>
 * 
 * Arguments:
 * <ul>
 *  <li> pbl: a Problem object,
 *  <li> map: an RCP to a Map object,
 *  <li> elementsList: an RCP to an ElementList object.
 * </ul>
 */
template <typename scalar>
SourcesList<scalar>::SourcesList(Problem &pbl, Teuchos::RCP<Map> map, Teuchos::RCP<ElementsList> elementsList)
{
    typedef local_ordinal_type local_ordinal_type;
    typedef global_ordinal_type global_ordinal_type;

    sources_number = pbl.Sources.size();
    local_ordinal_type maxNumElemsPerSource = 0;

    std::vector<std::vector<global_ordinal_type>> SourceElems = {};

    for (auto i = 0; i < sources_number; ++i)
    {
        std::vector<global_ordinal_type> mySourceElems = {};

        for (auto j = 0; j < pbl.Sources[i]->tag->elems.size(); ++j)
        {
            local_ordinal_type e = map->mapElems->getLocalElement(pbl.Sources[i]->tag->elems[j]->no - 1);

            if (elementsList->getElementType(e) == static_cast<int>(tbox::ELTYPE::TETRA4))
                ;
            else if (elementsList->getElementType(e) == static_cast<int>(tbox::ELTYPE::HEX8))
                ;
            else
                continue;

            mySourceElems.push_back(e);
        }

        SourceElems.push_back(mySourceElems);
        if (maxNumElemsPerSource < mySourceElems.size())
            maxNumElemsPerSource = mySourceElems.size();
    }

    sources_size = 0;
    sources_elements = 1;
    sources_elementsList = Kokkos::View<int **, Kokkos::LayoutRight>("R", sources_number, sources_elements + maxNumElemsPerSource);

    for (auto i = 0; i < sources_number; ++i)
    {
        sources_elementsList(i, sources_size) = SourceElems[i].size();

        for (auto j = 0; j < SourceElems[i].size(); ++j)
            sources_elementsList(i, sources_elements + j) = SourceElems[i][j];
    }

    sources_time = 0;
    sources_values = 1;
    sources_values_list = Kokkos::View<scalar ***, Kokkos::LayoutRight>("R", sources_number, 1, 2);
    for (auto i = 0; i < sources_number; ++i)
    {
        sources_values_list(i, 0, sources_time) = 0.; //to change latter

        sources_values_list(i, 0, sources_values) = pbl.Sources[i]->value;
    }
}
}; // namespace katropton

#endif //KATROPTON_SOURCESLIST_H
