#ifndef KATROPTON_RANDOMFIELD_H
#define KATROPTON_RANDOMFIELD_H

#include "Stokhos_Sacado_Kokkos_MP_Vector.hpp"
#include "Stokhos_KL_ExponentialRandomField.hpp"
#include "EnsembleTraits.h"
#include "wSpectralApproach.h"

template <typename Scalar, typename Device>
class RandomField
{
public:
    typedef Kokkos::View<Scalar *, Kokkos::LayoutLeft, Device> RandomVariableView;
    typedef typename RandomVariableView::size_type size_type;

    typedef typename Teuchos::ScalarTraits<Scalar>::coordinateType MeshScalar;
    typedef Stokhos::KL::ExponentialRandomField<MeshScalar, Device> rf_type;
    typedef Teuchos::ScalarTraits<MeshScalar> MST;
    typedef typename EnsembleTraits<Scalar>::value_type mean_type;

    bool isRandom = false;
    bool isExpRandom = true;
    rf_type m_rf;            // Exponential random field
    RandomVariableView m_rv; // KL random variables
    mean_type g;             // Mean
    mean_type delta;         // Dispersion level
    mean_type a, b;
    int num_rv;
    int ndim = 3; // Dimension of the physical domain
    Teuchos::RCP<SpectralApproach<Scalar, Device>> sa;

    RandomField(Teuchos::RCP<Teuchos::ParameterList> randomParams, Kokkos::View<Scalar *, Kokkos::LayoutLeft, Device> _m_rv)
    {
        isRandom = randomParams->get("Is random", false);
        if (isRandom)
        {
            g = randomParams->get("Mean", MST::one());
            delta = randomParams->get("Dispersion level", MST::zero());
            a = log(g / (sqrt(1 + delta * delta)));
            b = sqrt(log(1 + delta * delta));

            isExpRandom = randomParams->get("Is exp", true);
            if (isExpRandom)
            {
                const MeshScalar BeginX = randomParams->get("Begin X", MST::zero());
                const MeshScalar BeginY = randomParams->get("Begin Y", MST::zero());
                const MeshScalar BeginZ = randomParams->get("Begin Z", MST::zero());
                const MeshScalar EndX = randomParams->get("End X", MST::one());
                const MeshScalar EndY = randomParams->get("End Y", MST::one());
                const MeshScalar EndZ = randomParams->get("End Z", MST::one());
                num_rv = randomParams->get("Number random variables", 0);
                const MeshScalar correlation_lengthX = randomParams->get("Correlation length X", MST::one());
                const MeshScalar correlation_lengthY = randomParams->get("Correlation length Y", MST::one());
                const MeshScalar correlation_lengthZ = randomParams->get("Correlation length Z", MST::one());

                if (correlation_lengthZ == 0)
                    ndim = 2;

                const int num_KL_X = randomParams->get("Number of KL Terms X", 0);
                const int num_KL_Y = randomParams->get("Number of KL Terms Y", 0);
                const int num_KL_Z = randomParams->get("Number of KL Terms Z", 0);

                mean_type eps = randomParams->get("Bound Perturbation Size", 1e-6);
                mean_type tol = randomParams->get("Nonlinear Solver Tolerance", 1e-10);
                int max_it = randomParams->get("Maximum Nonlinear Solver Iterations", 100);

                Teuchos::ParameterList solverParams;
                solverParams.set("Number of KL Terms", num_rv);
                solverParams.set("Mean", MST::zero());
                solverParams.set("Standard Deviation", MST::one());
                solverParams.set("Bound Perturbation Size", eps);
                solverParams.set("Nonlinear Solver Tolerance", tol);
                solverParams.set("Maximum Nonlinear Solver Iterations", max_it);
                Teuchos::Array<MeshScalar> domain_upper(ndim, 1.0),
                    domain_lower(ndim, 0.0),
                    correlation_lengths(ndim, correlation_lengthX);
                Teuchos::Array<int> num_KL_per_dim(ndim, 1.0);

                domain_upper[0] = EndX;
                domain_upper[1] = EndY;
                if (ndim == 3)
                    domain_upper[2] = EndZ;

                domain_lower[0] = BeginX;
                domain_lower[1] = BeginY;
                if (ndim == 3)
                    domain_lower[2] = BeginZ;

                correlation_lengths[0] = correlation_lengthX;
                correlation_lengths[1] = correlation_lengthY;
                if (ndim == 3)
                    correlation_lengths[2] = correlation_lengthZ;

                if (num_KL_X != 0 && num_KL_Y != 0 && num_KL_Z != 0)
                {
                    num_KL_per_dim[0] = num_KL_X;
                    num_KL_per_dim[1] = num_KL_Y;
                    if (ndim == 3)
                        num_KL_per_dim[2] = num_KL_Z;
                    solverParams.set("Number of KL Terms per dimension", num_KL_per_dim);
                }

                solverParams.set("Domain Upper Bounds", domain_upper);
                solverParams.set("Domain Lower Bounds", domain_lower);
                solverParams.set("Correlation Lengths", correlation_lengths);

                m_rf = rf_type(solverParams);
                m_rv = _m_rv;

                std::ofstream outputfile_solverParams;
                std::ofstream outputfile_m_rv;

                outputfile_solverParams.open("RandomFieldParams.txt");
                outputfile_m_rv.open("RandomVariables.txt");

                outputfile_solverParams << solverParams << std::endl;

                for (size_t i = 0; i < m_rv.extent(0); ++i)
                    outputfile_m_rv << m_rv(i) << std::endl;

                outputfile_solverParams.close();
                outputfile_m_rv.close();
            }
            else
            {
                MeshScalar a = randomParams->get("Correlation length", MST::one());
                MeshScalar r = randomParams->get("Wavenumber cutoff", MST::one());
                size_t muw = randomParams->get("Wavenumber discretization points", 40);
                bool gauss = randomParams->get("Gaussian field", true);

                unsigned long seed_Z = randomParams->get("Z seed", 42);
                unsigned long seed_Phi = randomParams->get("Phi seed", 186);

                size_t wait_Z = randomParams->get("Number Z of previously drawn samples", 0);
                size_t wait_Phi = randomParams->get("Number Phi of previously drawn samples", 0);

                Kokkos::View<Scalar **, Kokkos::LayoutLeft, Device> Z;
                Kokkos::View<Scalar **, Kokkos::LayoutLeft, Device> Phi;

                Z = construct_Z<Scalar, Device>(muw, seed_Z, wait_Z);
                Phi = construct_Phi<Scalar, Device>(muw, seed_Phi, wait_Phi);

                sa = Teuchos::rcp(new SpectralApproach<Scalar, Device>(Z, Phi, a, r, muw, gauss));
            }
        }
    }

    KOKKOS_INLINE_FUNCTION
    Scalar operator()(const MeshScalar x, const MeshScalar y, const MeshScalar z) const
    {
        if (isRandom)
        {
            Scalar Xi = 0.;
            if (isExpRandom)
            {
                if (ndim == 3)
                {
                    const MeshScalar point[3] = {x, y, z};
                    Xi = m_rf.evaluate(point, m_rv);
                }
                else
                {
                    const MeshScalar point[2] = {x, y};
                    Xi = m_rf.evaluate(point, m_rv);
                }
            }
            else
                Xi = sa->operator()(x, y);

            return exp(a + b * Xi);
        }
        else
            return Teuchos::ScalarTraits<Scalar>::one();
    }
};

#endif //KATROPTON_RANDOMFIELD_H
