#include "wWeight.h"
#include "wProblem.h"
#include "wTag.h"
#include "wElement.h"
#include "wNode.h"
#include <algorithm>
using namespace katropton;

/**
 * @brief Weight constructor
 * 
 * Arguments:
 * <ul>
 *  <li> pbl: a Problem object,
 *  <li> name: the name of the weighted domain of the msh,
 *  <li> _which_dof: a vector that specify which degree(s) of freedom are weighted,
 *  <li> _values: a vector that specify the values of the weights
 *  (the length of _which_dof and _values must be the same),
 *  <li> ensemble_size: the ensemble size.
 * </ul>
 */
Weight::Weight(katropton::Problem &pbl,
               std::string const &name,
               std::vector<int> _which_dof,
               std::vector<double> _values,
               size_t ensemble_size) : Group(pbl.msh, name), which_dof(_which_dof), values(_values)
{
    for (auto e : tag->elems)
        for (auto n : e->nodes)
            nodes.push_back(n);

    x_values = {};
    y_values = {};
    z_values = {};
    T_values = {};
    for (auto i = 0; i < ensemble_size; ++i)
    {
        x_values.push_back(values[0]);
        y_values.push_back(values[1]);
        z_values.push_back(values[2]);
        T_values.push_back(values[3]);
    }
    pbl.Weights.push_back(this);
}

/**
 * @brief Weight constructor
 * 
 * Arguments:
 * <ul>
 *  <li> pbl: a Problem object,
 *  <li> name: the name of the weighted domain of the msh,
 *  <li> x: a int which specify whether the x displacement component is weighted,
 *  <li> v_x: value of the weight,
 *  <li> y: a int which specify whether the y displacement component is weighted,
 *  <li> v_y: value of the weight,
 *  <li> z: a int which specify whether the z displacement component is weighted,
 *  <li> v_z: value of the weight,
 *  <li> T: a int which specify whether the temperature is constrained,
 *  <li> v_T: value of the weight,
 *  <li> ensemble_size: the ensemble size.
 * </ul>
 */
Weight::Weight(Problem &pbl,
               std::string const &name,
               int x,
               double v_x,
               int y,
               double v_y,
               int z,
               double v_z,
               int T,
               double v_T,
               size_t ensemble_size) : Group(pbl.msh, name)
{
    for (auto e : tag->elems)
        for (auto n : e->nodes)
            nodes.push_back(n);

    which_dof = {};
    which_dof.push_back(x);
    which_dof.push_back(y);
    which_dof.push_back(z);
    which_dof.push_back(T);

    values = {};
    values.push_back(v_x);
    values.push_back(v_y);
    values.push_back(v_z);
    values.push_back(v_T);

    x_values = {};
    y_values = {};
    z_values = {};
    T_values = {};
    for (auto i = 0; i < ensemble_size; ++i)
    {
        x_values.push_back(v_x);
        y_values.push_back(v_y);
        z_values.push_back(v_z);
        T_values.push_back(v_T);
    }

    pbl.Weights.push_back(this);
}

/**
 * @brief Weight constructor
 * 
 * Arguments:
 * <ul>
 *  <li> pbl: a Problem object,
 *  <li> name: the name of the weighted domain of the msh,
 *  <li> x: a int which specify whether the x displacement component is weighted,
 *  <li> v_x: vector of values of the weight (one per sample),
 *  <li> y: a int which specify whether the y displacement component is weighted,
 *  <li> v_y: vector of values of the weight (one per sample),
 *  <li> z: a int which specify whether the z displacement component is weighted,
 *  <li> v_z: vector of values of the weight (one per sample),
 *  <li> T: a int which specify whether the temperature is constrained,
 *  <li> v_T: vector of values of the weight (one per sample),
 *  <li> ensemble_size: the ensemble size.
 * </ul>
 */
Weight::Weight(Problem &pbl,
               std::string const &name,
               int x,
               std::vector<double> v_x,
               int y,
               std::vector<double> v_y,
               int z,
               std::vector<double> v_z,
               int T,
               std::vector<double> v_T) : Group(pbl.msh, name)
{
    for (auto e : tag->elems)
        for (auto n : e->nodes)
            nodes.push_back(n);

    which_dof = {};
    which_dof.push_back(x);
    which_dof.push_back(y);
    which_dof.push_back(z);
    which_dof.push_back(T);

    values = {};
    values.push_back(v_x[0]);
    values.push_back(v_y[0]);
    values.push_back(v_z[0]);
    values.push_back(v_T[0]);

    x_values = {};
    y_values = {};
    z_values = {};
    T_values = {};
    for (auto i = 0; i < v_x.size(); ++i)
        x_values.push_back(v_x[i]);

    for (auto i = 0; i < v_y.size(); ++i)
        y_values.push_back(v_y[i]);

    for (auto i = 0; i < v_z.size(); ++i)
        z_values.push_back(v_z[i]);

    for (auto i = 0; i < v_T.size(); ++i)
        T_values.push_back(v_T[i]);

    pbl.Weights.push_back(this);
}

/**
 * @brief Display information.
 */
void Weight::write(std::ostream &out) const
{
    out << "Weights on " << *tag << "\n";
}
