#!/usr/bin/env python
# -*- coding: latin-1; -*-

from __future__ import print_function
from builtins import str
from builtins import range
import katropton as m
import tbox
import tbox.gmsh as gmsh
import katropton.utilities as u
from PyTrilinos import Teuchos
import shlex
import subprocess
import os
import numpy as np
import tarfile
import glob
from fwk.wutils import parseargs


def tar_results(directory, filename, rank):
    os.chdir(directory)
    if rank == 0:
        tf = tarfile.open(filename + ".tar.gz", mode="w:gz")
        for file in glob.glob('*'):
            if file.endswith(".tar.gz"):
                print(file + ' is not included')
            else:
                tf.add(file)
        tf.close()


def model(msh,
          E,
          p,
          ensemble_size,
          case_name,
          zero_initial_guess=True,
          prec=False):
    comm, rank, size = m.utilities.import_MPI()

    pbl = m.Problem(msh, comm)

    nu = 0.2 * np.ones(ensemble_size)
    k = 0.114
    d = 0.

    zero = np.zeros(ensemble_size)
    m.Medium(pbl, "Body 1", "test", E, nu, k, d)

    m.Dirichlet(pbl, "Surf 1 1", "Clamped", 1, 0., 1, 0., 1, 0., 1, 0.,
                ensemble_size)
    m.Neumann(pbl, "Surf 1 2", "heat fluxes", 0, zero, 0, zero, 1, p, 0, zero,
              ensemble_size)

    solverList = Teuchos.ParameterList()
    solverList['Convergence Tolerance'] = 10**(-8)
    solverList['Maximum Iterations'] = 100
    solverList['Verbosity'] = 33
    solverList['Flexible Gmres'] = False
    solverList['Output Style'] = 1
    solverList['Output Frequency'] = 1

    if prec:
        mueluParams = Teuchos.ParameterList()
        mueluParams['number of equations'] = 3
        mueluParams['max levels'] = 1
        mueluParams['coarse: type'] = "RELAXATION"

        mueluParamsSub2 = Teuchos.ParameterList()
        mueluParamsSub2['relaxation: type'] = "Jacobi"
        mueluParamsSub2['relaxation: sweeps'] = 1
        mueluParamsSub2['relaxation: damping factor'] = 1.

        mueluParams['verbosity'] = "low"

        mueluParams['coarse: params'] = mueluParamsSub2

        solverList['mueluParams'] = mueluParams
        solverList['Use preconditioner'] = True
    else:
        solverList['Use preconditioner'] = False

    if not zero_initial_guess:
        file_dir = os.path.dirname(__file__)
        solverList['Initial guess'] = file_dir + '/initial_guess.txt'

    solverList['type'] = "BlockGmres"
    solverList['Maximum active set iteration'] = 1

    solverList['Print Teuchos timers'] = True
    solverList['Write txt files'] = True
    solverList['Write matrix and vectors'] = True

    slv = m.IterativeSolver(pbl, solverList, 3, ensemble_size)

    if not os.path.exists(case_name):
        os.makedirs(case_name)
    os.chdir(case_name)

    if rank == 0:
        np.savetxt('E.txt', E)
        np.savetxt('p.txt', p)

    slv.start()

    os.chdir('..')


def main():
    comm, rank, size = u.import_MPI()

    useER = m.UseEnsembleReduction()

    geo_name = '1_cube.geo'
    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    msh, mesh_name, partitioned_mesh_name = u.read_mesh_MPI(
        geo_name, file_dir, work_dir, comm, rank, size)

    E_mean = 1.
    E_min = 0.75
    E_max = 1.25

    E_min_2 = 0.95
    E_max_2 = 1.05

    E_min_3 = 1.2
    E_max_3 = 1.3

    p_mean = 0.1
    p_min = 0.075
    p_max = 0.125

    ensemble_size = 1
    E = E_mean * np.ones(ensemble_size)
    p = p_mean * np.ones(ensemble_size)
    directory_case_id = 'mean'

    nThreads = u.Initialize_Kokkos()
    model(msh, E, p, ensemble_size, directory_case_id)

    ensemble_sizes = [8, 16, 32]

    for ensemble_size in ensemble_sizes:
        E = np.linspace(E_min, E_max, ensemble_size)
        p = p_mean * np.ones(ensemble_size)
        directory_case_id = 'test_1_s_' + str(ensemble_size)

        model(msh, E, p, ensemble_size, directory_case_id)

        E = np.linspace(E_min_2, E_max_2, ensemble_size)
        p = p_mean * np.ones(ensemble_size)
        directory_case_id = 'test_6_s_' + str(ensemble_size)

        model(msh, E, p, ensemble_size, directory_case_id)

        E = np.linspace(E_min_3, E_max_3, ensemble_size)
        p = p_mean * np.ones(ensemble_size)
        directory_case_id = 'test_7_s_' + str(ensemble_size)

        model(msh, E, p, ensemble_size, directory_case_id)

        E = E_mean * np.ones(ensemble_size)
        p = np.linspace(p_min, p_max, ensemble_size)
        directory_case_id = 'test_2_s_' + str(ensemble_size)

        model(msh, E, p, ensemble_size, directory_case_id)

        E = np.linspace(E_min, E_max, ensemble_size)
        p = np.linspace(p_min, p_max, ensemble_size)
        directory_case_id = 'test_3_s_' + str(ensemble_size)

        model(msh, E, p, ensemble_size, directory_case_id)

        E = np.linspace(E_min, E_max, ensemble_size)
        p = np.linspace(p_min, p_max, ensemble_size)
        directory_case_id = 'test_4_s_' + str(ensemble_size)

        model(msh, E, p, ensemble_size, directory_case_id, True, True)

        E = np.linspace(E_min, E_max, ensemble_size)
        p = np.linspace(p_min, p_max, ensemble_size)
        directory_case_id = 'test_5_s_' + str(ensemble_size)

        model(msh, E, p, ensemble_size, directory_case_id, False, False)

    u.Finalize_Kokkos()

    if useER:
        tar_results(work_dir, 'all_results_ER', rank)
    else:
        tar_results(work_dir, 'all_results_NER', rank)


if __name__ == "__main__":
    main()
