/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wGmshExport.h"
#include "wTag.h"
#include "wElement.h"
#include "wNode.h"
#include "wResults.h"

using namespace tbox;

GmshExport::GmshExport(std::shared_ptr<MshData> _msh) : MshExport(_msh)
{
}

GmshExport::~GmshExport()
{
    std::cout << "~GmshExport()\n";
}

/**
 * @brief Save the mesh on disk
 */
void GmshExport::save(std::string const &fname) const
{
    std::string wname = fname + ".msh";
    std::cout << "writing file: " << wname << "... " << std::flush;

    // write gmsh .msh file with stdio commands
    FILE *file = fopen(wname.c_str(), "wt");
    if (!file)
        throw std::runtime_error("Error opening file (save)");

    // write format
    fprintf(file, "$MeshFormat\n");
    fprintf(file, "2.2 %d %d\n", 0, 8);
    fprintf(file, "$EndMeshFormat\n");

    // write physical names
    fprintf(file, "$PhysicalNames\n");
    std::vector<Tag *> named;
    for (auto &p : msh->ptags)
        if (p.second->name != Tag::defaultName())
            named.push_back(p.second);
    fprintf(file, "%lu\n", static_cast<unsigned long>(named.size()));
    for (auto t : named)
        fprintf(file, "%d %d \"%s\"\n", t->dim, t->no, t->name.c_str());
    fprintf(file, "$EndPhysicalNames\n");

    // nodes
    fprintf(file, "$Nodes\n");
    fprintf(file, "%lu\n", static_cast<unsigned long>(msh->nodes.size()));
    for (auto n : msh->nodes)
        fprintf(file, "%d %.16g %.16g %.16g\n", n->no, n->pos(0), n->pos(1), n->pos(2));
    fprintf(file, "$EndNodes\n");

    // elements
    fprintf(file, "$Elements\n");
    fprintf(file, "%lu\n", static_cast<unsigned long>(msh->elems.size()));
    for (auto e : msh->elems)
    {
        size_t ntag = 2;
        if (!e->parts.empty())
            ntag += 1 + e->parts.size();
        fprintf(file, "%d %d %lu", e->no, static_cast<int>(e->type()), static_cast<unsigned long>(ntag));
        if (e->ptag)
            fprintf(file, " %d", e->ptag->no);
        else
            fprintf(file, " 0");
        if (e->etag)
            fprintf(file, " %d", e->etag->no);
        else
            fprintf(file, " 0");
        if (!e->parts.empty())
        {
            fprintf(file, " %lu", static_cast<unsigned long>(e->parts.size()));
            for (auto p : e->parts)
                fprintf(file, " %d", p->no);
        }
        for (auto n : e->nodes)
            fprintf(file, " %d", n->no);
        fprintf(file, "\n");
    }
    fprintf(file, "$EndElements\n");

    // close file
    fclose(file);
    std::cout << "done" << std::endl;
}

/**
 * @brief Save results (associated to a mesh) on disk
 */
void GmshExport::save(std::string const &fname, Results const &r) const
{
    std::string wname = fname + ".pos";
    std::cout << "writing file: " << wname << "... " << std::flush;

    FILE *file;
    if (binary)
        file = fopen(wname.c_str(), "wb");
    else
        file = fopen(wname.c_str(), "wt");
    if (!file)
        throw std::runtime_error("Error opening file (save)");

    // write format
    fprintf(file, "$MeshFormat\n");
    fprintf(file, "2.2 %d %d\n", binary ? 1 : 0, 8);
    if (binary)
    {
        int one = 1;
        fwrite(&one, sizeof(int), 1, file);
    }
    fprintf(file, "$EndMeshFormat\n");

    // node data: scalars_at_nodes
    for (auto &p : r.scalars_at_nodes)
    {
        fprintf(file, "$NodeData\n");
        fprintf(file, "%d\n", 1);                                              // nb of string tags
        fprintf(file, "\"%s\"\n", p.first.c_str());                            // the name of the view
        fprintf(file, "%d\n", 1);                                              // nb of real tags (1)
        fprintf(file, "%g.16\n", r.time);                                      //   1. the time value
        fprintf(file, "%d\n", 3);                                              // nb of integer tags (3)
        fprintf(file, "%d\n", r.nt);                                           //   1. time step index
        fprintf(file, "%d\n", 1);                                              //   2. nb of field component (1, 3 or 9)
        fprintf(file, "%lu\n", static_cast<unsigned long>(msh->nodes.size())); //   3. nb of entities in the view
        for (size_t ii = 0; ii < msh->nodes.size(); ++ii)
        {
            if (binary)
            {
                fwrite(&(msh->nodes[ii]->no), sizeof(int), 1, file);
                fwrite(&((*(p.second))[ii]), sizeof(double), 1, file);
            }
            else
                fprintf(file, "%d %.16g\n", msh->nodes[ii]->no, (*(p.second))[ii]);
        }
        fprintf(file, "$EndNodeData\n");
    }

    // element data: scalars_at_elems
    for (auto &p : r.scalars_at_elems)
    {
        fprintf(file, "$ElementData\n");
        fprintf(file, "%d\n", 1);                                              // nb of string tags
        fprintf(file, "\"%s\"\n", p.first.c_str());                            // the name of the view
        fprintf(file, "%d\n", 1);                                              // nb of real tags (1)
        fprintf(file, "%g.16\n", r.time);                                      //   1. the time value
        fprintf(file, "%d\n", 3);                                              // nb of integer tags (3)
        fprintf(file, "%d\n", r.nt);                                           //   1. time step index
        fprintf(file, "%d\n", 1);                                              //   2. nb of field component (1, 3 or 9)
        fprintf(file, "%lu\n", static_cast<unsigned long>(msh->elems.size())); //   3. nb of entities in the view
        for (size_t ii = 0; ii < msh->elems.size(); ++ii)
        {
            if (binary)
            {
                fwrite(&msh->elems[ii]->no, sizeof(int), 1, file);
                fwrite(&(*(p.second))[ii], sizeof(double), 1, file);
            }
            else
                fprintf(file, "%d %.16g\n", msh->elems[ii]->no, (*(p.second))[ii]);
        }
        fprintf(file, "$EndElementData\n");
    }

    // node data: vectors_at_nodes
    for (auto &p : r.vectors_at_nodes)
    {
        fprintf(file, "$NodeData\n");
        fprintf(file, "%d\n", 1);                                              // nb of string tags
        fprintf(file, "\"%s\"\n", p.first.c_str());                            // the name of the view
        fprintf(file, "%d\n", 1);                                              // nb of real tags (1)
        fprintf(file, "%g.16\n", r.time);                                      //   1. the time value
        fprintf(file, "%d\n", 3);                                              // nb of integer tags (3)
        fprintf(file, "%d\n", r.nt);                                           //   1. time step index
        fprintf(file, "%d\n", 3);                                              //   2. nb of field component (1, 3 or 9)
        fprintf(file, "%lu\n", static_cast<unsigned long>(msh->nodes.size())); //   3. nb of entities in the view
        for (size_t ii = 0; ii < msh->nodes.size(); ++ii)
        {
            Eigen::Vector3d const &v = (*(p.second))[ii];
            if (binary)
            {
                fwrite(&msh->nodes[ii]->no, sizeof(int), 1, file);
                fwrite(&(v(0)), sizeof(double), 3, file);
            }
            else
                fprintf(file, "%d %.16g %.16g %.16g\n", msh->nodes[ii]->no, v(0), v(1), v(2));
        }
        fprintf(file, "$EndNodeData\n");
    }

    // element data: vectors_at_elems
    for (auto &p : r.vectors_at_elems)
    {
        fprintf(file, "$ElementData\n");
        fprintf(file, "%d\n", 1);                                              // nb of string tags
        fprintf(file, "\"%s\"\n", p.first.c_str());                            // the name of the view
        fprintf(file, "%d\n", 1);                                              // nb of real tags (1)
        fprintf(file, "%g.16\n", r.time);                                      //   1. the time value
        fprintf(file, "%d\n", 3);                                              // nb of integer tags (3)
        fprintf(file, "%d\n", r.nt);                                           //   1. time step index
        fprintf(file, "%d\n", 3);                                              //   2. nb of field component (1, 3 or 9)
        fprintf(file, "%lu\n", static_cast<unsigned long>(msh->elems.size())); //   3. nb of entities in the view
        for (size_t ii = 0; ii < msh->elems.size(); ++ii)
        {
            Eigen::Vector3d const &v = (*(p.second))[ii];
            if (binary)
            {
                fwrite(&(msh->elems[ii]->no), sizeof(int), 1, file);
                fwrite(&(v(0)), sizeof(double), 3, file);
            }
            else
                fprintf(file, "%d %.16g %.16g %.16g\n", msh->elems[ii]->no, v(0), v(1), v(2));
        }
        fprintf(file, "$EndElementData\n");
    }

    // node data: tensors_at_nodes
    for (auto &p : r.tensors_at_nodes)
    {
        fprintf(file, "$NodeData\n");
        fprintf(file, "%d\n", 1);                                              // nb of string tags
        fprintf(file, "\"%s\"\n", p.first.c_str());                            // the name of the view
        fprintf(file, "%d\n", 1);                                              // nb of real tags (1)
        fprintf(file, "%g.16\n", r.time);                                      //   1. the time value
        fprintf(file, "%d\n", 3);                                              // nb of integer tags (3)
        fprintf(file, "%d\n", r.nt);                                           //   1. time step index
        fprintf(file, "%d\n", 9);                                              //   2. nb of field component (1, 3 or 9)
        fprintf(file, "%lu\n", static_cast<unsigned long>(msh->nodes.size())); //   3. nb of entities in the view
        for (size_t ii = 0; ii < msh->nodes.size(); ++ii)
        {
            Eigen::MatrixXd const &v = (*(p.second))[ii];
            if (binary)
            {
                fwrite(&msh->nodes[ii]->no, sizeof(int), 1, file);
                for (int i = 0; i < 3; ++i)
                    for (int j = 0; j < 3; ++j)
                        fwrite(&(v(i, j)), sizeof(double), 1, file);
            }
            else
                fprintf(file, "%d %.16g %.16g %.16g %.16g %.16g %.16g %.16g %.16g %.16g\n",
                        msh->nodes[ii]->no,
                        v(0, 0), v(0, 1), v(0, 2),
                        v(1, 0), v(1, 1), v(1, 2),
                        v(2, 0), v(2, 1), v(2, 2));
        }
        fprintf(file, "$EndNodeData\n");
    }

    // element data: tensors_at_elems
    for (auto &p : r.tensors_at_elems)
    {
        fprintf(file, "$ElementData\n");
        fprintf(file, "%d\n", 1);                                              // nb of string tags
        fprintf(file, "\"%s\"\n", p.first.c_str());                            // the name of the view
        fprintf(file, "%d\n", 1);                                              // nb of real tags (1)
        fprintf(file, "%g.16\n", r.time);                                      //   1. the time value
        fprintf(file, "%d\n", 3);                                              // nb of integer tags (3)
        fprintf(file, "%d\n", r.nt);                                           //   1. time step index
        fprintf(file, "%d\n", 9);                                              //   2. nb of field component (1, 3 or 9)
        fprintf(file, "%lu\n", static_cast<unsigned long>(msh->elems.size())); //   3. nb of entities in the view
        for (size_t ii = 0; ii < msh->elems.size(); ++ii)
        {
            Eigen::MatrixXd const &v = (*(p.second))[ii];
            if (binary)
            {
                fwrite(&(msh->elems[ii]->no), sizeof(int), 1, file);
                for (int i = 0; i < 3; ++i)
                    for (int j = 0; j < 3; ++j)
                        fwrite(&(v(i, j)), sizeof(double), 1, file);
            }
            else
                fprintf(file, "%d %.16g %.16g %.16g %.16g %.16g %.16g %.16g %.16g %.16g\n",
                        msh->elems[ii]->no,
                        v(0, 0), v(0, 1), v(0, 2),
                        v(1, 0), v(1, 1), v(1, 2),
                        v(2, 0), v(2, 1), v(2, 2));
        }
        fprintf(file, "$EndElementData\n");
    }

    // close file
    fclose(file);
    std::cout << "done" << std::endl;
}

void GmshExport::write(std::ostream &out) const
{
    out << "GmshExport on " << *msh << std::endl;
}