/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wTetra4.h"
#include "wSfTetra4.h"
#include "wGaussTetra4.h"
#include "wCacheTetra4.h"
#include "wMemTetra4.h"
#include "wNode.h"
#include "wFct0.h"

using namespace tbox;

Tetra4::Tetra4(int n, Tag *_ptag, Tag *_etag,
               std::vector<Tag *> &_parts, std::vector<Node *> &nods) : Element(n, _ptag, _etag, _parts, nods)
{
    pmem = new MemTetra4(*this, static_cast<int>(getCache().gauss.getN()));
}
Tetra4::~Tetra4()
{
    if (pmem)
        delete pmem;
}

/**
 * @brief Return private memory handling Jacobian for this Tetra4 element 
 */
Mem &Tetra4::getVMem()
{
    return getMem();
}

/**
 * @brief Return cache handling shape functions and Gauss points for all Tetra4 elements
 */
Cache &Tetra4::getVCache()
{
    return getCache();
}

/**
 * @brief Compute Jacobian Matrix at Gauss point k
 * 
 * J_ij(3,3) = diN_n * xj_n (n is the node)
 */
double Tetra4::buildJ(size_t k, Eigen::MatrixXd &J) const
{
    Eigen::MatrixXd const &dff = getCache().getDsf(k);

    J = Eigen::Matrix3d::Zero();
    size_t i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
        J += dff.col(i) * ((*it)->pos).transpose();
    return J.determinant();
}

/**
 * @brief Compute mass matrix
 * 
 * M_ij(4,4) = int N_i*N_j dV
 */
Eigen::MatrixXd Tetra4::buildM()
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    // Gauss integration
    Eigen::MatrixXd M = Eigen::Matrix4d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Shape functions
        Eigen::VectorXd const &ff = cache.getSf(k);
        M += ff * ff.transpose() * gauss.getW(k) * mem.getDetJ(k);
    }
    return M;
}

/**
 * @brief Compute stiffness matrix for scalar field
 * 
 * K_ij(4,4) = int f*dN_j*dN_i dV
 */
Eigen::MatrixXd Tetra4::buildK(std::vector<double> const &u, Fct0 const &fct)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    // Gauss integration
    Eigen::MatrixXd K = Eigen::Matrix4d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Jacobian inverse and shape functions
        Eigen::Matrix3d const &J = mem.getJinv(k);
        Eigen::MatrixXd const &dff = cache.getDsf(k);

        // Elementary stiffness matrix
        K += (J * dff).transpose() * J * dff * fct.eval(*this, u, k) * gauss.getW(k) * mem.getDetJ(k);
    }
    return K;
}

/**
 * @brief Compute (first part of) derivative stiffness matrix for scalar field (Newton method)
 * 
 * K_ij(4,4) = int (df du*dN_j du*dN_i) dV
 */
Eigen::MatrixXd Tetra4::buildDK(std::vector<double> const &u, Fct0 const &fct)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    // Gauss integration
    Eigen::MatrixXd K = Eigen::Matrix4d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Jacobian inverse, shape functions and gradient
        Eigen::Matrix3d const &J = mem.getJinv(k);
        Eigen::MatrixXd const &dff = cache.getDsf(k);
        Eigen::VectorXd gradu = computeGrad(u, k);

        // Elementary stiffness matrix
        K += (gradu.transpose() * J * dff).transpose() * gradu.transpose() * J * dff * fct.evalD(*this, u, k) * gauss.getW(k) * mem.getDetJ(k);
    }
    return K;
}

/**
 * @brief Compute derivative of flux vector for a scalar field (Newton method)
 * 
 * s_i(4) = int f*du*dN_i dV
 */
Eigen::VectorXd Tetra4::buildDs(std::vector<double> const &u, Fct0 const &f)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    // Gauss integration
    Eigen::VectorXd s = Eigen::Vector4d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
        s += computeGrad(u, k).transpose() * mem.getJinv(k) * cache.getDsf(k) * f.eval(*this, u, k) * gauss.getW(k) * mem.getDetJ(k); // elementary flux vector
    return s;
}

/**
 * @brief Compute linear elasticity stiffness matrix
 * 
 * K_ij(12,12) = int eps_k H_ijkl eps_l dV = B(12,6)H(6,6)B(6,12)
 */
Eigen::MatrixXd Tetra4::buildK_mechanical(Eigen::MatrixXd const &H)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    Eigen::MatrixXd K = Eigen::MatrixXd::Zero(12, 12);
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Jacobian inverse
        Eigen::Matrix3d const &invJ = mem.getJinv(k);
        // Fill B matrix (shape functions)
        Eigen::MatrixXd const &dff = cache.getDsf(k);
        Eigen::MatrixXd B = Eigen::MatrixXd::Zero(6, 3 * nodes.size());
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            // Compute [Ni,x Ni,y, Ni_z]
            Eigen::Vector3d dN = invJ * dff.col(i);
            B(0, 3 * i) = dN(0);
            B(1, 3 * i + 1) = dN(1);
            B(2, 3 * i + 2) = dN(2);
            B(3, 3 * i + 1) = dN(2);
            B(3, 3 * i + 2) = dN(1);
            B(4, 3 * i) = dN(2);
            B(4, 3 * i + 2) = dN(0);
            B(5, 3 * i) = dN(1);
            B(5, 3 * i + 1) = dN(0);
        }

        // Compute stiffness matrix
        K += B.transpose() * H * B * gauss.getW(k) * mem.getDetJ(k);
    }
    return K;
}

/* -------------------------------------------------------------------------- */

/**
 * @brief Compute thermomechanical mass matrix
 */
Eigen::MatrixXd Tetra4::buildS_thermomechanical(Eigen::MatrixXd const &D)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    // Gauss integration
    Eigen::MatrixXd S_e = Eigen::MatrixXd::Zero(12, 4);
    for (size_t a = 0; a < gauss.getN(); ++a)
    {
        // Jacobian and shape functions
        double detJ = mem.getDetJ(a);
        Eigen::Matrix3d const &J = mem.getJinv(a);
        Eigen::VectorXd const &ff = cache.getSf(a);
        Eigen::MatrixXd const &dff = cache.getDsf(a);

        for (size_t i = 0; i < nodes.size(); ++i)
        {
            Eigen::Vector3d factor = D * J * dff.col(i);
            for (size_t j = 0; j < nodes.size(); ++j)
                for (size_t k = 0; k < 3; ++k)
                    S_e(i * 3 + k, j) += factor(k) * ff(j) * gauss.getW(a) * detJ;
        }
    }
    return S_e;
}

/**
 * @brief Compute thermal matrix
 */
Eigen::MatrixXd Tetra4::buildL_thermal(Eigen::MatrixXd const &C)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    // Gauss integration
    Eigen::MatrixXd L_e = Eigen::Matrix4d::Zero();
    for (size_t a = 0; a < gauss.getN(); ++a)
    {
        // Jacobian
        double detJ = mem.getDetJ(a);
        Eigen::Matrix3d const &J = mem.getJinv(a);
        Eigen::MatrixXd const &dff = cache.getDsf(a);

        for (size_t i = 0; i < nodes.size(); ++i)
        {
            Eigen::Vector3d factor1 = C * J * dff.col(i);
            for (size_t j = i; j < nodes.size(); ++j)
            {
                Eigen::Vector3d factor2 = J * dff.col(j);
                L_e(i, j) += factor1.dot(factor2) * gauss.getW(a) * detJ;
            }
        }
    }
    for (size_t i = 0; i < nodes.size(); ++i)
        for (size_t j = i; j < nodes.size(); ++j)
            L_e(j, i) = L_e(i, j);
    return L_e;
}

/**
 * @brief Compute stress and strain
 */
void Tetra4::strain_stress(Eigen::MatrixXd &Strain, Eigen::MatrixXd &Stress, Eigen::MatrixXd const &H, std::vector<double> const &u)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    Strain = Stress = Eigen::Matrix3d::Zero();
    Eigen::Matrix3d Tmp = Eigen::Matrix3d::Zero();
    Eigen::MatrixXd u_e(3, nodes.size());
    for (size_t i = 0; i < nodes.size(); ++i)
    {
        u_e(0, i) = u[3 * (nodes[i]->row)];
        u_e(1, i) = u[3 * (nodes[i]->row) + 1];
        u_e(2, i) = u[3 * (nodes[i]->row) + 2];
    }
    for (size_t a = 0; a < gauss.getN(); ++a)
    {
        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        Eigen::Matrix3d const &J = mem.getJinv(a);
        Eigen::Matrix3d du_d_Phi = Eigen::Matrix3d::Zero();
        Eigen::MatrixXd const &dff = cache.getDsf(a);

        for (size_t i = 0; i < nodes.size(); ++i)
        {
            for (auto k = 0; k < 3; ++k)
                for (auto j = 0; j < 3; ++j)
                    du_d_Phi(j, k) += dff(j, i) * u_e(k, i);
        }

        for (size_t k = 0; k < 3; ++k)
            for (size_t m = 0; m < 3; ++m)
                for (size_t i = 0; i < 3; ++i)
                    Tmp(k, m) += J(k, i) * du_d_Phi(i, m) / (nodes.size());
    }
    for (size_t i = 0; i < 3; ++i)
        for (size_t j = 0; j < 3; ++j)
            Strain(i, j) = (Tmp(i, j) + Tmp(j, i)) / 2;

    for (size_t i = 0; i < 3; ++i)
        for (size_t j = 0; j < 3; ++j)
            for (size_t k = 0; k < 3; ++k)
                for (size_t l = 0; l < 3; ++l)
                    Stress(i, j) += H(i * 3 + j, k * 3 + l) * Strain(k, l);
}

/* -------------------------------------------------------------------------- */

/**
 * @brief Compute gradient of a scalar field at Gauss point k
 */
Eigen::VectorXd Tetra4::computeGrad(std::vector<double> const &u, size_t k)
{
    // Solution vector
    Eigen::Vector4d ue;
    for (size_t i = 0; i < nodes.size(); ++i)
        ue(i) = u[nodes[i]->row];
    // gradient in global axis: inv(J)_ij * d_jN_i * ue_i
    return getMem().getJinv(k) * getCache().getDsf(k) * ue;
}

/**
 * @brief Compute square of gradient of a scalar field at Gauss point k
 */
double Tetra4::computeGrad2(std::vector<double> const &u, size_t k)
{
    return computeGrad(u, k).squaredNorm();
}

/**
 * @brief Compute intergal of scalar function
 * 
 * I = int f dV
 */
double Tetra4::computeV(std::vector<double> const &u, Fct0 const &fct)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    // Gauss integration
    double V = 0.0;
    for (size_t k = 0; k < gauss.getN(); ++k)
        V += fct.eval(*this, u, k) * gauss.getW(k) * mem.getDetJ(k);
    return V;
}

/**
 * @brief Return the element volume
 */
double Tetra4::computeV()
{
    return getMem().getVol();
}
